/*
 * Copyright 2009-2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var serializer;

function getSerializer() {
    if( serializer == null ) serializer = new XMLSerializer();
    return serializer;
}

//----------------------------------------------------------------

function serializeDocument() {
    removeAllLinks();
    var scripts = removeScripts();
    var res = getSerializer().serializeToString( document );
    injectScripts( scripts );
    addLocationLinks();
    return res;
}

function serializeDocumentRaw() {
    return getSerializer().serializeToString( document );
}

function injectScripts( scripts ) {
    $( "head" ).append( scripts );
}

function removeScripts() {
    var res = $("#jquery_js").add("#edit_js").add("#edit_css");
    res.remove();
    return res;
}

/**
 * Mark intersections of range and text node range.
 * 
 * new:
 * 
 * <span class="requality_text id_${uuid}">
 *     <a name="${uuid}" id="id_${uuid}" class="requality_id"/>
 *     <span class="requality_link requality_link_(in)?visible">${id}</span>
 *     ...
 * </span>
 * <span class="requality_text id_${uuid}">...</span>
 * 
 * old:
 * 
 * <span class="location_text id_${uuid}"><a name="${uuid}" id="id_${uuid}" class="location_id id_${uuid}"/>...</span>
 * <span class="location_link id_${uuid} location_link_(in)?visible>${id}</span>
 * <span class="location_text id_${uuid}">...</span>
 * <span class="location_text id_${uuid}">...<span class="location_end id_${uuid}"/></span>
 * 
 * @param range    
 * @param textNode
 * @return 1 if intersection of range and textNode is not empty and is not equal to textNode
 * To mark such intersection we add one child node. 
 */
function markTextRange( id, range, textNode ) {
    var r = document.createRange();
    r.selectNode(textNode);
    var res = 0;
    var isStart = false;
    var isEnd = false;
    // r is before range
    if( range.compareBoundaryPoints(Range.END_TO_START,r) == 1 ) return 0;
    // r is after range
    if( range.compareBoundaryPoints(Range.START_TO_END,r) == -1 ) return 0;
    switch( range.compareBoundaryPoints(Range.START_TO_START,r) ) {
    case 1:
        r.setStart( range.startContainer, range.startOffset );
        res = 1;
    case 0:
    	isStart = true;
    }
    switch( range.compareBoundaryPoints(Range.END_TO_END,r) ) {
    case -1:
        r.setEnd( range.endContainer, range.endOffset );
        res = 1;
    case 0:
        isEnd = true;
    }
    if( r.toString().trim().length == 0 ) return 0;
    var reqwrap = $("<span/>").attr("class","requality_text id_"+id);
    r.surroundContents( reqwrap.get(0) );
    if( isStart )
    {
        reqwrap.prepend( $("<a/>").attr("name",id).attr("id","id_"+id).attr("class","requality_id") );
    }
//    if( isEnd )
//    {
//        reqwrap.append($("<span/>").attr("class","location_end id_"+id));
//    }
    return res;
}

function markRange( id, range, node ) {
    for( var i = 0; i < node.childNodes.length; i++ )
    {
        var child = node.childNodes.item(i);
        if( child.nodeType == Node.TEXT_NODE )
        {
            i += markTextRange( id, range, child );
        } else {
            markRange( id, range, child );
        }
    }
}

function markSelection( id ) {
    var selection = window.getSelection();
    if( selection.rangeCount > 1 ) alert( "Only the first part of selection will be marked" );
    if( selection.rangeCount == 0 ) {
    	alert( "Selection is empty" );
    	return null;
    }
    var range = selection.getRangeAt(0);
    var text = getSerializer().serializeToString( range.cloneContents() );
//    alert(text);
//    var text = range.toString();
    var container = range.commonAncestorContainer;
    if( container.nodeType == Node.TEXT_NODE )
    {
    	// range is a part of text node. Mark intersection. 
        markTextRange( id, range, container );
//        alert(container.parentNode.innerHTML);
    } else {
    	// Find text node children and mark intersections.
        markRange( id, range, container );
//        alert(container.innerHTML);
    }
    return text;
}

/**
 * This function wrap selection in special requality tags.
 * <br><br><i>Can be used in java</i>
 * @param id UUID
 * @param userId user visible id( <i>For example, 'some.id.01'</i> )
 * @return - selection text
 */
function createLocation( id, reqId ) {
    var text = markSelection( id );
    addLocationLink( id, reqId );
    return text;
}

function addLocationLinks() {
	var locations = getDocumentLocationIds();
	if( locations.length == 0 ) return;
	var id2links = getLocationsToRequirementMap( locations );
	for( var i = 0; i < id2links.length; i++ )
	{
		var chunk = id2links[i];
		for( var j = 1; j < chunk.length; j++ )
		{
			addLocationLink( chunk[0], chunk[j] );
		}
	}		
}

function addLocationLink( id, link ) {
	var v = areIdentifiersVisible();
    $("#id_"+id).after($("<span>"+link+"</span>").addClass("requality_link").addClass(v?"requality_link_visible":"requality_link_invisible"));
}

function removeLocationLink( id, link ) {
    $("#id_"+id).parent().children(".requality_link :contains('"+link+"')").remove();
}

function removeAllLinks() {
    $(".requality_link").remove();
}

function getSelectedLocationUUID() {
    var s = window.getSelection();
    var f = s.focusNode;
    var id = $(f).parents(".requality_text").attr("class");
    if( id == null )
    {
//    	alert( "No selected location" );
    	return null;
    }
    id = id.substring(18);
    return id;
}

function getDocumentLocationIds()
{
	var res = [];
	$(".requality_id").each(
		function()
		{
			var id = this.getAttribute("id");
			res.push( id.substring(3) );
		}
	);
	return res;
}

function removeLocation( id ) {
	var t = $("span.requality_text.id_" + id );
	t.children( ".requality_id" ).remove();
	t.children( ".requality_link" ).remove();
	t.each(
		function()
		{
	    	while( this.childNodes.length > 0 )
	    	{
		        this.parentNode.insertBefore( this.childNodes.item(0), this );
		    }
		}
	);
	t.remove();
}

/**
 * Selecting requality Location node and makes it orange.
 * <br><br><i>Can be used in java</i>
 */
function selectNode() {
    alert( 'selectNode' );
}

/**
 * STUB
 * @return true
 */
function areIdentifiersVisible() {
	return $(".requality_link").hasClass("requality_link_visible");
}

/**
 * Shows user visible ids in document
 * <br><br><i>Can be used in java</i> 
 */
function showIdentifiers() {
    $(".requality_link").removeClass("requality_link_invisible").addClass("requality_link_visible");
}

/**
 * Hides user visible ids in document
 * <br><br><i>Can be used in java</i>
 */
function hideIdentifiers() {
    $(".requality_link").removeClass("requality_link_visible").addClass("requality_link_invisible");
}

function getBrowserInfo() {
	var appName= navigator.appName;
    var appVersion= navigator.userAgent;
    return "Browser: " + appName + "; Version: " + appVersion;
}

/**
 * <i>Can be used in java</i>
 * @return - <b>true</b> if selection can be wrapped by DOM element<br>
 * - and <b>false</b> otherwise.   
 */
function isSelectable() {
//    alert( 'isSelectable' );
    return true;
}

/**
 * <i>Can be used in java</i>
 * @return - <b>true</b> if selected node can be unwrapped
 */
function isRemovable() {
    return getSelectedLocationUUID() != null;
}

document.onclick =
    function( e )
    {
        // right mouse button
        if( e.button == 2 )
        {
            var s = window.getSelection();
            var n = s.rangeCount;
            for( var i = 0; i < n; i++ )
            {
                var r = s.getRangeAt(i);
                // if one of selection ranges contains current mouse position - do nothing
                // WARNING: isPointInRange is mozilla specific function
                if( r.isPointInRange( e.rangeParent, e.rangeOffset ) ) return;
            }
            // collapse selection to current mouse position
            s.collapse( e.rangeParent, e.rangeOffset );
        }
    };

function propNames( obj )
{
    if( !obj ) return "obj is undefined";
    var s = "/*" + obj.toString() + ":*/{";
    var comma = false;
    for( p in obj ) { 
        if( comma ) s += ",";
        comma = true;
        s += "|" + p + "=" + obj[p];
    }
    s += "\n}";
    return s;
}

function inspect( obj ) {
    alert( propNames( obj ) );
}

function tree( node )
{
	var s = "";
	switch( node.nodeType )
	{
	case Node.ELEMENT_NODE:
		s += "<"+node.nodeName;
		s += " class=" + node.getAttribute("class") + ">";
		for( var i = 0; i < node.childNodes.length; i++ )
		{
			s += "\n"+tree( node.childNodes.item(i) );
		}
		s += "\n</"+node.nodeName+">";
		break;
	case Node.TEXT_NODE:
		s += node.nodeValue;
	}
	return s;
}	

addLocationLinks();
setInitialLocation();
